#    ANARCI - Antibody Numbering and Antigen Receptor ClassIfication
#    Copyright (C) 2015 Oxford Protein Informatics Group (OPIG)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.#
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
ANARCI - Antigen Receptor Numbering And ClassIfication

Oxford Protein Informatics Group (OPIG). 2015

ANARCI performs alignments of sequences to databases of Hidden Markov Models (HMMs).
Those that align with a significant score are classified by species and chain type.
They are then numbered with a scheme of the user's choosing. 

Currently implemented schemes: 
    IMGT

Currently recognisable species (chains):
    # TCR
    Human, mouse (alpha, beta, delta, gamma)
    # MHC
    Human, Mouse (g-alpha1, g-alpha2, g-alpha, g-beta)
    
Notes:
 o The AHo scheme is implemented heuristically.
 o Fragments of antibodies can be recognised. Play with the bit_score_threshold in
   _parse_hmmer_query if you want to investigate this feature.
 o Really long CDR3s will not be identified. although the v-gene segment will be.
 o Unusual N and C terminal residues are often not considered as part of the domain by
   HMMER. Thus, they are not included in the numbering.


'''

import os
import sys, re
import tempfile
from textwrap import wrap
from subprocess import Popen, PIPE
from itertools import groupby

# Import the HMMER parser from the distributed version of Biopython.
from tcrcdr.anarci.Bio.SearchIO.HmmerIO import Hmmer3TextParser as HMMERParser

# Import from the schemes submodule
from .schemes import *
#from mhc_alleles import mhc_alleles
from collections import defaultdict

#all_species = all_germlines['V']['B'].keys()

amino_acids = sorted(list("QWERTYIPASDFGHKLCVNM"))
set_amino_acids = set(amino_acids)
anarci_path  = os.path.split(__file__)[0]

# chain_type_to_class = {"A":"A", "B":"B", "G":"G", "D":"D", "GA": "GA", "GB2": "GB2", "GA1": "GA1", "GA2": "GA2", "C": "C" }

HMM_path =  os.path.join( anarci_path, "dat", "HMMs" )

all_reference_states = list(range( 1, 129)) # These are the IMGT reference states (matches)
b2m_reference_states = list(range( 1, 115)) # B2M/C reference state has 114 positions
mhc_reference_states = list(range( 1, 107)) # MHC reference states have 106 positions
cd1_reference_states = list(range( 1, 103)) # CD1 reference states have 102 positions

class HMMscanError(Exception):
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(HMMscanError, self).__init__(message)

## Utility functions ##
def read_fasta(filename):
    """
    Read a sequence file and parse as description, string 
    """
    return [ r for r in fasta_iter(filename) ]

def fasta_iter(fasta_name):
    """
    Given a fasta file. yield tuples of header, sequence
    https://www.biostars.org/p/710/
    """
    fh = open(fasta_name)
    faiter = (x[1] for x in groupby(fh, lambda line: line[0] == ">"))
    for header in faiter:
        header = header.next()[1:].strip()
        seq = "".join(s.strip() for s in next(faiter))
        yield header, seq


def write_fasta(sequences, f):
    """
    Write a list of sequences to file. 
    
    should be a list of name, sequence tuples
    
    f should be an open file
    """
    for name, sequence in sequences:
        print(">%s"%name, file=f)
        print('\n'.join(['\n'.join(wrap(block, width=80)) for block in sequence.splitlines()]), file=f)
    
    
def validate_sequence(sequence):
    """
    Check whether a sequence is a protein sequence or if someone has submitted something nasty.
    """
    assert len(sequence) < 10000, "Sequence too long."
    assert not (set( sequence.upper() ) - set_amino_acids ), "Unknown amino acid letter found in sequence: %s"% ", ".join(list((set( sequence.upper() ) - set_amino_acids )))    
    return True

def anarci_output(numbered, sequences, alignment_details, outfile, sequence_id=None, domain_id=None):
    """
    Output to open file

    If sequence_id is specified as an integer then only this sequence will be printed. 
    Otherwise all sequences will be printed.

    If domain_id is specified as an integer then only this domain will be printed. 
    Otherwise all domains will be printed.

    If domain_id is specified then sequence_id must also be specified. 
    """       
    assert (sequence_id is not None) or (sequence_id is None and domain_id is None), "If domain_id is specified, sequence_id must also be specified."

    for i in range(len(numbered)):
        if sequence_id is None:
            print("# %s"%sequences[i][0], file=outfile) # print the name
        if numbered[i] is not None:
            if sequence_id is not None:
                if i != sequence_id: continue
            print("# ANARCI numbered", file=outfile)
            for j in range( len(numbered[i])): # Iterate over domains
                if domain_id is not None:
                    if j != domain_id: continue
                print("# Domain %d of %d"%(j+1, len(numbered[i]) ), file=outfile)
                print("# Most significant HMM hit", file=outfile)
                print("#|species|chain_type|e-value|score|seqstart_index|seqend_index|", file=outfile)
                alignment_details[i][j]["evalue"] = str( alignment_details[i][j]["evalue"] )
                print("#|%s|%s|%s|%.1f|%d|%d|"%tuple( [alignment_details[i][j][field] for field in 
                                                                     ["species","chain_type","evalue","bitscore"]] 
                                                                   +[ numbered[i][j][1], numbered[i][j][2]] ), file=outfile)
                
                #if 'germlines' in alignment_details[i][j]:
                #    print >> outfile, '# Most sequence-identical germlines'
                #    print >> outfile, '#|species|v_gene|v_identity|j_gene|j_identity|'
                #    (species, vgene), vid =alignment_details[i][j]['germlines']['v_gene']
                #    if vgene is None:
                #        vgene, vid = 'unknown', 0
                #    (_,jgene), jid =alignment_details[i][j]['germlines']['j_gene']
                #    if jgene is None:
                #        jgene, jid = 'unknown', 0
                #    print >> outfile, '#|%s|%s|%.2f|%s|%.2f|'%(species, vgene, vid, jgene, jid )    

                # chain_type = chain_type_to_class[  alignment_details[i][j]["chain_type"] ]
                chain_type = alignment_details[i][j]["chain_type"]

                print("# Scheme = %s"%alignment_details[i][j]["scheme"], file=outfile)
                for (index, insertion), aa in numbered[i][j][0]:
                    print(chain_type, ("%d"%index).ljust(5), insertion, aa, file=outfile)
        print("//", file=outfile)


## Parsing and recognising domain hits from hmmscan ##

def _domains_are_same(dom1, dom2):
    """
    Check to see if the domains are overlapping.
    @param dom1: 
    @param dom2: 

    @return: True or False  
    """
    dom1, dom2 = sorted( [dom1, dom2], key=lambda x: x.query_start  )

    if dom2.query_start >= dom1.query_end:
        return False
    return True


def _parse_hmmer_query(query, bit_score_threshold = 90):
    """
    
    @param query: hmmer query object from TCRDB.Biopython
    @param bit_score_threshold: the threshold for which to consider a hit a hit. 
    
    The function will identify multiple domains if they have been found and provide the details for the best alignment for each domain.
    This allows the ability to identify single chain fvs and engineered antibody sequences as well as the capability in the future for identifying constant domains. 

    """
    hit_table = [ ['id', 'description', 'evalue', 'bitscore', 'bias', 
                    'query_start', 'query_end' ] ]

    # Find the best hit for each domain in the sequence.

    top_descriptions, domains,state_vectors = [], [], []
    prev_id = set()

    if query.hsps: # We have some hits
        for hsp in sorted(query.hsps, key=lambda x: x.evalue): # Iterate over the matches of the domains in order of their e-value (most significant first)
            new=True
            if hsp.bitscore >= bit_score_threshold: # Only look at those with hits that are over the threshold bit-score.
                for i in range( len(domains) ): # Check to see if we already have seen the domain
                    hit_domain = hsp.hit_id.split("_")[-1]

                    if _domains_are_same( domains[i], hsp ):
                        if hit_domain in prev_id:
                            new = False
                        elif hit_domain in ["A","D"]:
                            break

                    prev_id.add(hit_domain)

                # Some TRAV genes can be TRDV too, and we need multiple state vectors in these cases.
                prev_id.add( hsp.hit_id.split("_")[-1] )

                if new: # It is a new domain and this is the best hit. Add it for further processing.
                    domains.append( hsp )
                    hit_table.append( [ hsp.hit_id, hsp.hit_description, hsp.evalue, hsp.bitscore, hsp.bias, hsp.query_start, hsp.query_end] )
                    top_descriptions.append(  dict( list(zip(hit_table[0], hit_table[-1])) ) ) # Add the last added to the descriptions list. 

        # Reorder the domains according to the order they appear in the sequence.         
        ordering = sorted( list(range(len(domains))), key=lambda x: domains[x].query_start)
        domains = [ domains[_] for _ in ordering ]
        top_descriptions = [ top_descriptions[_] for _ in ordering ]         

    for i in range(len(domains)): # If any significant hits were identified parse and align them to the reference state.
        domains[i].order = i
        species, chain = top_descriptions[i]["id"].split("_")
        state_vectors.append( _hmm_alignment_to_states(domains[i], chain) ) # Alignment to the reference states.
        top_descriptions[i][ "species"] = species # Reparse
        top_descriptions[i][ "chain_type"] = chain
        top_descriptions[i][ "query_start"] = state_vectors[-1][0][-1] # Make sure the query_start agree if it was changed

    return hit_table, state_vectors, top_descriptions


def _hmm_alignment_to_states(hsp, chain):
    """
    Take a hit hsp and turn the alignment into a state vector with sequence indices
    """

    # Extract the strings for the reference states and the posterior probability strings     
    reference_string = hsp.aln_annotation["RF"]
    state_string = hsp.aln_annotation["PP"]
    
    assert len(reference_string) == len(state_string), "Aligned reference and state strings had different lengths. Don't know how to handle"
    
    # Extract the start an end points of the hmm states and the sequence
    # These are python indices i.e list[ start:end ] and therefore start will be one less than in the text file
    _hmm_start = hsp.hit_start
    _hmm_end = hsp.hit_end
     
    _seq_start = hsp.query_start
    _seq_end = hsp.query_end

    # Handle cases where there are n terminal modifications.
    # In most cases the user is going to want these included in the numbered domain even though they are not 'antibody like' and 
    # not matched to the germline. Only allow up to a maximum of 5 unmatched states at the start of the domain
    # Adds a bug here if there is a very short linker between a scfv domains with a modified n-term second domain
    # Thus this is only done for the first identified domain ( hence order attribute on hsp )
    if hsp.order == 0 and _hmm_start and _hmm_start < 5: 
        n_extend = _hmm_start 
        if _hmm_start > _seq_start:
            n_extend = min( _seq_start , _hmm_start - _seq_start )
        state_string = '8'*n_extend + state_string  
        reference_string = 'x'*n_extend + reference_string
        _seq_start = _seq_start - n_extend
        _hmm_start = _hmm_start - n_extend

    # Generate lists for the states and the sequence indices that are included in this alignment
    # The reference state depends on chain type
    if chain in ("GA", "GB", "GA1", "GA2", "GA2L"):
        hmm_states = mhc_reference_states[ _hmm_start : _hmm_end ] 
    elif chain in ("GA1L", "MR1", "MR2"):
        hmm_states = cd1_reference_states[ _hmm_start : _hmm_end ] 
    elif chain == "C":
        hmm_states = b2m_reference_states[ _hmm_start : _hmm_end ] 
    else:
        hmm_states = all_reference_states[ _hmm_start : _hmm_end ] 

    sequence_indices = list(range(_seq_start,  _seq_end))
    h, s = 0, 0 # initialise the current index in the hmm and the sequence
    
    state_vector = []
    # iterate over the state string (or the reference string)
    for i in range( len(state_string) ):
        if reference_string[i] == "x": # match state
            state_type = "m"
        else: # insert state
            state_type = "i"
        
        if state_string[i] == ".": # overloading if deleted relative to reference. delete_state
            state_type = "d"
            sequence_index = None
        else:
            sequence_index = sequence_indices[s]    
        # Store the alignment as the state identifier (uncorrected IMGT annotation) and the index of the sequence
        
        state_vector.append(  ((hmm_states[h], state_type),  sequence_index )  )        

        # Updates to the indices         
        if state_type == "m":
            h+=1
            s+=1
        elif state_type == "i":
            s+=1
        else: # delete state
            h+=1
        
    return state_vector


def parse_hmmer_output(filedescriptor="", bit_score_threshold=90):
    """
    Parse the output of HMMscan and return top alignment and the score table for each input sequence.
    """
    results  = []
    if type(filedescriptor) is str:
        openfile = open
    elif type(filedescriptor) is int:
        openfile = os.fdopen
    
    with openfile(filedescriptor) as inputfile:
        p = HMMERParser( inputfile )
        for query in p:
            results.append(_parse_hmmer_query(query,bit_score_threshold=bit_score_threshold ))
    return results


def run_hmmer(sequence_list,hmm_database="ALL",hmmerpath="", ncpu=None, bit_score_threshold=90):
    """
    Run the sequences in sequence list against a precompiled hmm_database.

    Those sequence that have a significant hit with a bit score over a threshold will
    be recognised and an alignment given. The alignment will be used to number the 
    sequence.

    @param sequence_list: a list of (name, sequence) tuples. Both are strings
    @param hmm_database: The hmm database to use. Currently, all hmms are in the ALL database.
                         The code to develop new models is in build_pipeline in the git repo.
    @param hmmerpath: The path to hmmer binaries if not in the path
    @param ncpu: The number of cpu's to allow hmmer to use.
    """

    # Check that hmm_database is available
    
    assert hmm_database in ["ALL"], "Unknown HMM database %s"%hmm_database    
    HMM = os.path.join( HMM_path, "%s.hmm"%hmm_database )

    # Create a fasta file for all the sequences. Label them with their sequence index
    # This will go to a temp file
    fasta_filehandle, fasta_filename =  tempfile.mkstemp( ".fasta", text=True )
    with os.fdopen(fasta_filehandle,'w') as outfile:
        write_fasta(sequence_list, outfile)

    output_filehandle, output_filename =  tempfile.mkstemp( ".txt", text=True )

    # Run hmmer as a subprocess
    if hmmerpath:
        hmmscan = os.path.join(hmmerpath,"hmmscan")
    else:
        hmmscan = "hmmscan"
    try:
        if ncpu is None:
            command = [ hmmscan, "-o", output_filename, HMM,  fasta_filename]
        else:
            command = [ hmmscan, "-o", output_filename, "--cpu", str(ncpu), HMM,  fasta_filename]
        process = Popen( command, stdout=PIPE, stderr=PIPE  )
        _, pr_stderr = process.communicate()

        if pr_stderr:
            _f = os.fdopen(output_filehandle) # This is to remove the filedescriptor from the os. I have had problems with it before.
            _f.close()
            
            raise HMMscanError(pr_stderr)
        results = parse_hmmer_output(output_filehandle, bit_score_threshold=bit_score_threshold)
        
    finally:
        # clear up
        os.remove(fasta_filename)
        os.remove(output_filename)
        
    return results


def number_sequence_from_alignment(state_vector, sequence, chain_type=None):
    """
    Given you have an alignment. Give back the numbering
    
    @param state_vector: List of states from the hmm. Effectively these are imgt columns but CDR3 has not been redone. 
    @param sequence: The original sequence string or list.
    @param scheme: The numbering scheme to apply
    @param chain_type: The type of chain to apply numbering for.
                       This can be one of several options:
                       M/m     -> GA1/GA2 domains (MHC1)
                       N/n     -> GA/GB domains   (MHC2)
                       B/A/D/G -> TCR chains
    
    @return: A list of numbering identifier / amino acids tuples over the domain that has been numbered. The indices of the start (inclusive) and end point (exclusive) in the sequence for the numbering 
    """
    if chain_type in ("GA1", "GA2", "GA", "GB"):
        return number_mhc(state_vector, sequence)
    elif chain_type in ("GA1L", "GA2L", "MR1", "MR2"):
        return number_cd1(state_vector, sequence)
    elif chain_type == "B2M":
        return number_b2m(state_vector, sequence)
    else:
        return number_imgt(state_vector, sequence)
'''
def get_identity( state_sequence, germline_sequence, len_reference_state = 128 ):
    """
    Get the partially matched sequence identity between two aligned sequences. 
    Partial in the sense that gaps can be in the state_sequence.
    """
    # Ensure that the sequences are the expected length
    assert len( state_sequence) == len(germline_sequence ) == len_reference_state
    n, m = 0, 0
    for i in xrange( len_reference_state ):
        if germline_sequence[i] == "-":continue
        if state_sequence[i] == germline_sequence[i]: m+=1
        n+=1

    if not n:
        return 0    
    return float(m)/n
'''
def validate_numbering(numbering):
    """
    Wrapper to do some basic validation of the numbering.
    Further validation could be done but at the moment we just check that the numbering indices are incremental (they should be)
    """
    len_numbering = len(numbering)

    for i in range(1, len_numbering):
        if not numbering[i][0][0] >= numbering[i-1][0][0]:
            return False
    return True
'''
def run_germline_assignment(state_vector, sequence, chain_type, allowed_species=None ):
    """
    Find the closest sequence identity match.
    """
    genes={'v_gene': [None,None],
           'j_gene': [None,None]}
    
    # Iterate over the v-germline sequences of the chain type of interest.
    # The maximum sequence identity is used to assign the germline 
    if chain_type in all_germlines["V"]:

        # Extract the positions that correspond to match (germline) states. 
        state_dict = dict( ((i, 'm'),None) for i in xrange(1,129))
        state_dict.update(dict(state_vector))
        state_sequence = "".join([ sequence[state_dict[(i, 'm')]] if state_dict[(i,'m')] is not None else "-" for i in xrange(1,129) ])

        if allowed_species is not None:
            assert all( [ sp in all_germlines['V'][chain_type] for sp in allowed_species ] ), 'Unknown species specified. Choose from %s'%','.join( all_species )
        else:
            allowed_species = all_species
        seq_ids = {}
        for species in allowed_species:
            if species not in all_germlines["V"][ chain_type ]: continue # Previously bug.
            for gene, germline_sequence in all_germlines["V"][ chain_type ][ species ].iteritems():
                seq_ids[ (species, gene) ] = get_identity( state_sequence , germline_sequence )
        genes['v_gene' ][0] = max( seq_ids, key=lambda x: seq_ids[x] )
        genes['v_gene' ][1] = seq_ids[ genes['v_gene' ][0] ]
        
        # Use the assigned species for the v-gene for the j-gene. 
        # This assumption may affect exotically engineered abs but in general is fair.
        species = genes['v_gene' ][0][0]
        if chain_type in all_germlines["J"]:
            if species in all_germlines["J"][chain_type]:
                seq_ids = {}
                for gene, germline_sequence in all_germlines["J"][ chain_type ][ species ].iteritems():
                    seq_ids[ (species, gene) ] = get_identity( state_sequence , germline_sequence )
                genes['j_gene' ][0] = max( seq_ids, key=lambda x: seq_ids[x] )
                genes['j_gene' ][1] = seq_ids[ genes['j_gene' ][0] ]
     
    # If we have an MHC chain, then the reference state has to change 
    elif chain_type in ("GA1", "GA2", "GA", "GB", "GA1L", "GA2L", "MR1", "MR2"):

        if chain_type in ("GA1", "GA2", "GA", "GB"):
            end_chain = 107
        elif chain_type in ("GA1L", "MR1"):
            end_chain = 100
        else:
            end_chain = 103

        # Extract the positions that correspond to match (germline) states. 
        state_dict = dict( ((i, 'm'),None) for i in xrange(1, end_chain))
        state_dict.update(dict(state_vector))
        state_sequence = "".join([ sequence[state_dict[(i, 'm')]] if state_dict[(i,'m')] is not None else "-" for i in xrange(1, end_chain) ])

        seq_ids = {}
        genes = {chain_type: [None,None]}

        for species in ("human", "mouse"):
            for allele, seq in mhc_alleles[chain_type][species].iteritems():
                seq_ids[ (species, allele) ] = get_identity(state_sequence, seq, len_reference_state = end_chain-1)

        # MHCs are a bit unusual that we can have the same allele NAME but the sequence identity for just one of their G-Like domains can be >=99%
        max_identity = max(seq_ids.values())
        allele_subset = [ k for k in seq_ids if seq_ids[k] >= max_identity ]

        if len(allele_subset) == 1 and chain_type in ("MR1", "MR2", "GA1L", "GA2L"):
            genes[chain_type][0] = allele_subset[0]
        else:
            genes[chain_type][0] = allele_subset

        genes[chain_type][1] = max_identity

    # for b2m domains
    else:
        # Extract the positions that correspond to match (germline) states. 
        state_dict = dict( ((i, 'm'),None) for i in xrange(1,115))
        state_dict.update(dict(state_vector))
        state_sequence = "".join([ sequence[state_dict[(i, 'm')]] if state_dict[(i,'m')] is not None else "-" for i in xrange(1,115) ])

        seq_ids = {}
        genes = {chain_type: [None,None]}

        for species in ("human", "mouse"):
            for allele, seq in mhc_alleles[chain_type][species].iteritems():
                seq_ids[ (species, allele) ] = get_identity(state_sequence, seq, len_reference_state = 114)

        max_identity = max(seq_ids.values())
        # MHCs are a bit unusual that we can have the same allele NAME but the sequence identity for just one of their C-Like domains can be >=99%
        genes[chain_type][0] = sorted([ k for k in seq_ids if seq_ids[k] >= max_identity ], key = lambda z: int(z[1].split("*")[1]))[0] # Get the lowest numbered allele
        genes[chain_type][1] =  max_identity
        
    return genes
'''


##################################
# High level numbering functions #
##################################

# Main function for ANARCI
def anarci(sequences, database="ALL", output=False, outfile=None, allow=set(["B","A","D","G","GA1","GA2","GA","GB","B2M","GA1L","GA2L","MR1","MR2"]), hmmerpath="", ncpu=None, assign_germline=False, allowed_species=None, bit_score_threshold=90, correct_alpha = True):
    """
    The main function for anarci. This can be used to identify domains, number them and classify their receptor type (and species).
    Multiple sequences can be handed to the function at once. 
    For a more basic interface to the algorithm see the "number" function                
    
    @param sequences: A list of name and amino acid sequence string tuples. 
                      e.g. [ ("seq1","GSSH... ..."),
                             ("seq2","GSSH... ...") ]
    @param database:  The HMMER database that should be used. Normally not changed unless a custom db is created.
    @param output:    Boolean flag to say whether the result should be output.
    @param outfile:   The name of the file to output to. If output is True and outfile is None then output is printed
                      to stdout.
    @param allow:     A set containing the chain types that should be recognised. If chothia, kabat or martin is used
                      as the scheme, anarci will ignore tcr chains.
    @param hmmerpath: The path to hmmscan. If left unspecified then the PATH will be searched. 
    @param ncpu:      The number of cpu's that hmmer should be allowed to use. If not specified then the hmmscan 
                      default is used.
    @param assign_germline: Currently turned off for TCR CDR canonical form assignment
    @param allowed_species:
    @param bit_score_threshold: 90
    @param correct_alpha: Correct vague genes that can be used as a delta or alpha receptor.

    @return: Three lists. Numbered, Alignment_details and Hit_tables.
             Each list is in the same order as the input sequences list.
             A description of each entry in the three lists is as followed.
               o Numbered: will be None if no domain was found for that sequence or a list of domains with their 
                           numbering, start and finish indices.
               o Alignment_details: will be None if no domain was found for that sequence or a dictionary for each
                           domain identified containing the details of the alignment (chain type, e-value, species etc).
               o Hit_tables: None if no domain was found for that sequence or a nested list for each domain containing
                           the hit table from hmmscan.
    
    """
        
    # Perform the alignments of the sequences to the hmm database
    alignments = run_hmmer( sequences, hmm_database=database, hmmerpath=hmmerpath, ncpu=ncpu, bit_score_threshold=bit_score_threshold )    
    
    # Iteration over the sequence alignments performing the desired numbering 
    numbered = []
    alignment_details = []
    hit_tables = []

    mhc_count = defaultdict(int) # Counter to be used for tracking how many MHC domains we have

    for i in range(len(sequences)):
        # Unpack
        hit_table, state_vectors, detailss = alignments[i] # We may have multiple domains per sequence (e.g. single chain TCRs).

        # Iterate over all the domains in the sequence that have been recognised (typcially only 1 with the current hmms available)
        hit_numbered, hit_details = [], []
        for di in range( len( state_vectors ) ):
            state_vector = state_vectors[di]
            details      = detailss[di]
            details["scheme"]="imgt"
            details["query_name"]=sequences[i][0]

            if state_vector and details["chain_type"] in allow: # only number things that are allowed. We still keep the alignment details and hit_table
                try:
                    # Do the numbering and validate (for development purposes)
                    numbering, start, end = number_sequence_from_alignment(state_vector, sequences[i][1], chain_type=details["chain_type"])

                    # Validate the numbering being an increment; if the domain doesn't follow this rule, continue/don't number it.
                    if not validate_numbering(numbering):
                        continue

                    # Renumber GA2 (m) chains by adding 1000 as done in IMGT. This also applies for the GA2-Like domain in CD1.
                    # GB (n) chains are usually annotated separately.
                    if "GA2" in details['chain_type'] or details["chain_type"] == "MR2":
                        for _ in range(len(numbering)):
                            t = numbering[_]
                            numbering[_] = ((t[0][0]+1000, t[0][1]), t[1])
                    '''
                    # Some delta/alpha genes V genes can be used to form a TCR, so use the J gene germline identity to determine if it is a D or A receptor.
                    # http://atlasgeneticsoncology.org/Genes/TCRAID39.html
                    elif correct_alpha and details['chain_type'] in ["D", "A"]:
                        germ_details = run_germline_assignment( state_vector, sequences[i][1], details["chain_type"], allowed_species=allowed_species)
                        # If the detected germline is a versatile one, only then pursue correction.
                        if "/" in germ_details['v_gene'][0][1]:
                            # Get the current J-gene identity
                            current_j_id = germ_details['j_gene'][1]
                            # Test the opposing chain type
                            other_ch = "D" if details['chain_type'] == "A" else "A" 
                            new_j_id = run_germline_assignment( state_vector, sequences[i][1], other_ch, allowed_species=allowed_species)['j_gene'][1]

                            # If the new J gene has a higher ID then it's likely going to be that opposing chain instead.
                            if (new_j_id > current_j_id):
                                continue # Iterate to the next state vector instead.

                                #numbering, start, end = number_sequence_from_alignment(state_vector, sequences[i][1], chain_type = other_ch)
                                #current_org = details['species']
                                #for t in hit_table:
                                #    if t[0] == "%s_%s" % (current_org, other_ch):
                                #        details['chain_type'] = other_ch
                                #        details['bias'], details['query_start'], details['query_end'] = t[-3], t[-2], t[-1]
                                #        details['evalue'], details['bitscore'] = t[2], t[3]
                                #        details['id'] = t[0]
                    '''
                                
                    hit_numbered.append( (numbering, start, end) )
                    '''
                    if assign_germline:

                        germ_details = run_germline_assignment( state_vector, sequences[i][1], details["chain_type"], allowed_species=allowed_species)
                        details["germlines"] = germ_details

                        if details['chain_type'] in ("GA1","GA2"):
                            mhc_count[i] += 1

                        elif details['chain_type'] in ("GA", "GB"):
                            allelenames, percentage = details['germlines'][details['chain_type']]
                            details['germlines'] = {details['chain_type']: [ allelenames[0] , percentage ]}

                            if len( allelenames ) == 1:
                               details['germlines'] = {details['chain_type']: [ allelenames[0] , percentage ]}
                            else:
                               # Get the lowest allele by heuristic. The run_germline_assignment code already gets the germline with the highest ID.
                               try:
                                   lowallele = sorted( allelenames, key = lambda x: int(re.findall(r"HLA-\w+.(\d+)", x[1])[0]) )[0]
                               except IndexError:
                                   lowallele = sorted( allelenames, key = lambda x: x[1] )[0]

                               lowallele_index = allelenames.index(lowallele)
                               details['germlines'] = {details['chain_type']: [ (lowallele[0], lowallele[1]), percentage ]}
                    '''
                    # Query end seems to go up by 1? subtract 1 to keep indexing.
                    details['query_end'] -= 1
                    hit_details.append( details )

                except AssertionError as e: # Handle errors. Those I have implemented should be assertion.
                    print(str(e), file=sys.stderr)
                    sys.exit(1) # Validation went wrong. Error message will go to stderr. Want this to be fatal during development.

                except Exception as e:
                    print("Error: Something really went wrong that has not been handled", file=sys.stderr)
                    print(str(e), file=sys.stderr)
                    sys.exit(1)

        if hit_numbered: 
            numbered.append( hit_numbered )
            alignment_details.append( hit_details )
        else: 
            numbered.append( None )
            alignment_details.append( None )
        hit_tables.append(hit_table)

        # Do some post-processing to get the correct allele for the MHC
        if mhc_count[i] == 1:
            domain   = alignment_details[i][0]['chain_type']
            #germline = alignment_details[i][0]['germlines'][domain]
            #alignment_details[i][0]['germlines'] = {domain: [ germline[0][0], germline[1] ]}

        elif mhc_count[i] >= 2:
            # In case we have more than 2 but have e.g. 3 domains detected, just use the first two
            mhc_doms = mhc_count[i] % 2
            if mhc_doms == 0:
                mhc_doms = mhc_count[i]
            else:
                mhc_doms = mhc_count[i] - 1

            # Allele dictionary will be used to track which germline a MHC G domain belongs to
            '''
            mhc_allele_dict = {"GA1": [], "GA2": []}
            
            # Have a counter for the number of separate query sequences, and the number of separate domains detected per sequence
            for k, deets in enumerate(alignment_details[i]):
                # Within each deets variable, there are X many alignments corresponding to the number of domains found
                chtype = deets['chain_type']
                if chtype == "B2M":
                    continue

                mhc_allele_dict[chtype].append( (deets["germlines"][chtype], k) )

            # The code is a bit messy but basic procedure is to first see what alleles have been annotated for 
            # a particular domain that we have. e.g. GA1 -- since this can be attributed to multiple alleles, we have to 
            # also see which alleles have been attributed to the opposing domain (e.g. GA2 for GA1). Then use the consensus
            # allele for both domains. If there are multiple common alleles, we use the allele with the lowest number as a heuristic

            opposing_domain = {"GA1":"GA2", "GA2":"GA1"}

            # Iterate through GA and GA1 and pair up
            for domain in ("GA1", "GA2"):
                opposing = opposing_domain[domain]
                for j in xrange(len(mhc_allele_dict[domain])):
                    domain_alleles = mhc_allele_dict[domain][i][0][0]
                    opposing_alleles = mhc_allele_dict[opposing][i][0][0]
                    common_allele = list(set(domain_alleles) & set(opposing_alleles))

                    #assert len(common_allele) == 1, "We should ideally only have one common allele"
                    if len(common_allele) > 1:
                        # Sort alleles by the number of their allele group; e.g. HLA-A*0201 > HLA-A*0921 
                        common_allele = sorted(common_allele, key = lambda x: int( re.split(r"HLA-\w+.(\d+)",x[1])[1] ))[0]
                    elif len(common_allele) == 1:
                        common_allele = common_allele[0]
                    # If no common allele can be detected, list the 
                    else:
                        domain_allele    = sorted(domain_alleles, key = lambda x: int( re.split(r"HLA-\w+.(\d+)",x[1])[1] ))[0]
                        # opposing_allele  = sorted(opposing_alleles, key = lambda x: int( re.split(r"HLA-\w+.(\d+)",x[1])[1] ))[0]
                        mth_domain = mhc_allele_dict[domain][i][1]
                        allele_idx = domain_alleles.index(domain_allele)    
                        #alignment_details[i][mth_domain]['germlines'] = {domain: [ domain_allele, mhc_allele_dict[domain][j][0][1] ]}
                        break

                    # Use the allele_idx to get the seqID of the MHC, and the nth seq and mth domain index to annotate germlines
                    allele_idx = domain_alleles.index(common_allele)
                    mth_domain = mhc_allele_dict[domain][i][1]
                    #alignment_details[i][mth_domain]["germlines"] = {domain: [common_allele, mhc_allele_dict[domain][j][0][1]  ]}
            '''
        # A verbose error statement if necessary
        # if print_err:
        #     print >> sys.stderr, "Multiple possible alleles with %.2f%% sequence identity have been detected." % (100.*mhc_allele_dict[domain][i][0][1][allele_idx])
        #     print >> sys.stderr, "By heuristic, the allele with the lowest number allele group and polymorphisms was used."
    

    if output: 
        outto, close=sys.stdout, False
        if outfile:
            outto, close = open(outfile,'w'), True
        anarci_output(numbered, sequences, alignment_details, outto)
        if close:
            outto.close()

    return numbered, alignment_details, hit_tables
                

# Wrapper function for ANARCI for back compatibility with sabdab annotate functions. 
def number(sequence, database="ALL", allow=set(["B","A","D","G","GA1","GA2","GA","GB","B2M","GA1L","GA2L","MR1","MR2"])):
    """
    Given a sequence string, use anarci to number it using the scheme of choice.
    Only the first domain will be recognised and numbered.

    @param sequence: An amino acid sequence string
    @param scheme: The numbering scheme that should be applied. Choose from imgt, chothia, kabat or martin
    @param database: The HMMER database that should be used. Normally not changed unless a custom db is created.
    @param allow: A set containing the chain types that should be recognised. If chothia, kabat or martin is used
                  as the scheme, anarci will ignore tcr chains.

    @return: If the sequence can be numbered, the numbering, chain type.
    """
    validate_sequence( sequence )      
    
    if len(sequence) < 70: # Length check. ANARCI can number fragments of chains well. Encourage full domain numbering. 
        return False, False

    try:
        numbered, alignment_details, _ = anarci( [("sequence_0", sequence)], database=database, output=False, allow=allow )
    except AssertionError: # Catch where the user has tried to number a TCR with an antibody scheme
        return False, False

    ### TODO
    # We return the numbering list and the chain type where kappa and lambda chains are both "L" for light
    if numbered[0] and len(numbered[0]) == 1:
        seq_chain_type  = alignment_details[0][0]['chain_type']
        return numbered[0][0][0], seq_chain_type

    elif numbered[0] and len(numbered[0]) > 1 and alignment_details[0][0]['chain_type'] == "GA1":
        seq_chain_type = "MH1"
        numbered = numbered[0][0][0] + numbered[0][1][0]
        return numbered, seq_chain_type

    elif numbered[0] and len(numbered[0]) > 1 and alignment_details[0][0]['chain_type'] == "GA1L":
        seq_chain_type = "CD1"
        return [numbered[0][0][0], numbered[0][1][0]], seq_chain_type

    elif numbered[0] and len(numbered[0]) > 1 and alignment_details[0][0]['chain_type'] == "MR1":
        seq_chain_type = "MR1"
        return [numbered[0][0][0], numbered[0][1][0]], seq_chain_type

    # elif numbered[0] and len(numbered[0]) > 1 and alignment_details[0][0]['chain_type'] in ["B", "A", "D", "G"]:
    elif numbered[0] and len(numbered[0]) > 1:
        # Single Chain T-cell Receptor (e.g. 1bwm); we only gave one sequence yet two domains were detected!
        seq_chain_type = alignment_details[0][0]['chain_type'] + alignment_details[0][1]['chain_type']
        return [numbered[0][0][0], numbered[0][1][0]], seq_chain_type

    else:
        return False, False
